function [t,cellfractions,molecularconc,G] = HumanLiverRegen_para_optimizaton_liverfailure()
% Author:       Babita Verma
% Date:         01/09/2018
%
% Description:  This code solves the system of equations govenrning liver
%               regeneration and optimizes all the 33 parameters of the model using elastic net 
%               with intiial guess as mentioned in Table A1. 
%
% output:   Generates the regeneration profile and optimized parameters of a patient from Yamamoto et al.,"Prediction of postoperative
%           liver regeneration from clinical information using a data-led mathematical model"; that showed liver failure 
%
% Returns variables:    t: time in days post resection
%                       cellfractions:  the fraction of quiescent(Q),priming(P), and replicating
%                                       cells(R); matrix whose columns represent Q,P,R and rows denote time
%                                       point
%                       molecularconc:  the concentration of IE genes, GF,ECM,IL6, JAK, STAT3, SOCS3;
%                                       matrix whose columns represent molecular concentration and rows
%                                       time
%                       G:              relative cell mass; Vector
%
% Citation: Babita K. Verma, Pushpavanam Subramaniam, Rajanikanth Vadigepalli,
%           "Modeling the dynamics of human liver failure post 
%            liver resection"


%% Main code begins

global data data_t x0 tStart tEnd k lamda1 lamda2 k0_ch W eps Nss

% Time series data of patient ID135 from Yamamoto et al.,"Prediction of postoperative
% liver regeneration from clinical information using a data-led mathematical
% model", Scientific Reports, 2016
data = [0.48392555,0.48392555,0.450084602,0.349125776];
data_t = [0,1,7,30];

%define fraction of remnant liver mass
initialFraction = data(1);

%constant that accounts for senescent
eps = 0.01;
%initial steady state of liver mass
Nss = 0.99;

%defining weighs for Weighed-least-square, one for equally weighed 
w_v = ones(1,length(data_t));
w_v(3) = 1;
w_v(end) = 1; %weighing last data point
W = diag(w_v);

% Define model parameters
k = zeros(41,1);

% Molecular parameters
k(1) = initialFraction; % Fraction of liver mass remaining (N)
% k0_guess(2) = 16.8; % Metabolic load (M)
k0_guess(1) = 10.041128; % Metabolic load (M) k(2)
k0_guess(2) = 1.5; % TNF production k(3)
k0_guess(3) = 0.9; % TNF mRNA degrad k(4)
k0_guess(4) = 2e4; % JAK activation rate (V_JAK)
k0_guess(5) = 1e4; % Km JAK
k0_guess(6) = 0.4; % JAK degrad
k0_guess(7) = 2; % Conc. of monomeric STAT3
k0_guess(8) = 7.5e2; % STAT3 activation rate
k0_guess(9) = 0.4; % Km STAT3
k0_guess(10) = 0.1; % STAT3 degrad
k0_guess(11) = 2.4e4; % SOCS activation rate
k0_guess(12) = 7e-4; % Km SOCS3
k0_guess(13) = 0.4; % SOCS3 degrad.
k0_guess(14) = 1.5e-2; % SOCS3 inhibition constant
k0_guess(15) = 2.5e2; % IE response gene expression rate
k0_guess(16) = 18; % Km for IE genes
k0_guess(17) = 5; % IE gene degrad
k0_guess(18) = 7; % ECM degrad by MMPs
k0_guess(19) = 33; % ECM degrad

% Cellular parameters
k0_guess(20) = 0.113; % GF production
k0_guess(21) = 0.23; % GF degrad
k0_guess(22) = 6e-2; % k_up - GF uptake/production rate by ECM
k0_guess(23) = 7e-3; % k_Q
k0_guess(24) = 4.4e-3; % k_P
k0_guess(25) = 5.4e-2; % k_R
k0_guess(26) = 2e-2; % k_prol - specifies length of mitotic cycle
k0_guess(27) = 0.1; % k_req - requiescence rate
k0_guess(28) = 8; % theta_req (for sigma_req)
k0_guess(29) = 3; % beta_req (for sigma_req)
k0_guess(30) = 1e-1; %1e-2; % k_cd - cell death rate
% k0_guess(32) = 9e-3; % theta_cd
k0_guess(31)=0.011050;
k0_guess(32) = 0.026046; % beta_cd
k0_guess(33)=6.5675e-4;% growth constant for cell mass k(41)

%list of the names of all the parameters being optimized
seq = {'M' 'k_{IL6}' '\kappa_{IL6}' 'V_{JAK}' 'K^{JAK}_M' '\kappa_{JAK}' '[proSTAT3]'...
'V_{ST3}' 'K^{ST3}_M' '\kappa_{ST3}' 'V_{SOCS3}' 'K^{SOCS3}_M' '\kappa_{SOCS3}' ...
'K^{SOCS3}_I' 'V_{IE}' 'K^{IE}_M' '\kappa_{IE}' 'k_{deg}' '\kappa_{ECM}' 'k_{GF}' ...
'\kappa_{GF}' 'k_{up}' 'k_{QP}' 'k_{PR}' 'k_{RQ}' 'k_{prol}' 'k_{req}' '\theta_{req}' ...
'\beta_{req}' 'k_{cd}' '\theta_{cd}' '\beta_{cd}' 'k_G'};

%parameter range
lb = [0.58 0.15 0.09 2e3 1e3 0.04 0.01 7.5e1 0.04 0.05 1.4e4 4e-4 0.1 0.5e-2 1.5e2 15 3 5 30 0.05 0.1 4e-2 5e-3 3e-3 4e-2 1e-2 0.05 6 2 0.5e-1 0.03 3e-3 3e-4];
ub = [58 15 9 2e5 1e5 4 20 7.5e3 40 1 3.4e4 1e-3 0.7 2.5e-2 3.5e2 21 7 9 36 0.2 0.35 8e-2 9e-3 6e-3 6.5e-2 3e-2 0.15 10 4 1.5e-1 0.044 6e-3 7e-4];

%Weighage for Ridge and Lasso
lamda1 = 0.001;
lamda2 = 0.001;

%defining characteristic parameter as initial guess
k0_ch = k0_guess;

% Set initial hepatocyte states, cellfractions = (Q,P,R)
% and concentrations, molecularconc = [IE,GF,ECM,TNF,JAK,STAT3,SOCS3]
cellfractions = [k(1),0,0];
molecularconc = ones(1,7);
G0 = 1;
x0 = [cellfractions,molecularconc,G0]';

tStart = 0;
tEnd = 910*24; %~2.5 years

%call error function
AA = [];bb = [];AAeq = [];bbeq = [];
display('inside fmincon...')
[k0_opt,fval] = fmincon(@error_cal,k0_guess,AA,bb,AAeq,bbeq,lb,ub);
display('outside fmincon...')

%storing optimized parameters
k(2:33) = k0_opt(1:32);
k(41) = k0_opt(33);

% Steady-State parameters
k(34) = -k(3)*k(2) + k(5)/(1+k(6)) + k(4); %-22.3; % TNF production
k(35) = -k(5)/(1+k(6)) + k(7); %-1.6; % JAK production
k(36) = -k(9)*k(8)^2/(k(8)^2+k(10)*(1+1/k(15))) + ...
    k(16)/(1+k(17)) + k(12)/(1+k(13)) + k(11); %2.39e4; % STAT3 production
k(37) = -k(12)/(1+k(13)) + k(14); %2.4e4; % SOCS production
k(38) = -k(16)/(1+k(17)) + k(18); %-8.16; % IE gene production
k(39) = k(19) + k(20); %40; % ECM production
k(40) = -k(21)*k(2)/(Nss+eps) + k(23) + k(22); %-1.6; % GF production

% Call ODE solver
[t,x] = ode15s(@(t,x)LiverRegenODE(t,x,k,x0), [tStart:.1:tEnd], x0);

% Remove negative values not used in calculations
x(x(:,1) < 0,1) = 0;
x(x(:,2) < 0,2) = 0;
x(x(:,3) < 0,3) = 0;

% Organize results
cellfractions = x(:,1:3);
molecularconc = x(:,4:10);
G = x(:,11);

%scaled optimized parameter
K0_sc = k0_opt./k0_ch;
k_all = [k0_ch;k0_opt;K0_sc];
t = t/24; % time in days 

%sorting the parameter values 
[para ind] = sort(K0_sc,'descend');


%regeneration profile
figure;hold on;
plot(data_t,data,'ro','MarkerFacecolor','r');
plot(t,cellfractions(:,1)+G.*(cellfractions(:,2)+cellfractions(:,3)),'r','linewidth',2);
set(gca,'FontSize',12);
xlabel('Time post-PHx (days)');
ylabel('Fraction of Liver');
ylim([0 1.10000001]); xlim([0 732]);
title('Regeneration Profile','FontSize',18);
box on;legend('Yamamoto et al.','Optmized');
hold off;

%scaled optimized parameters
figure;hold on;
plot(ones(1,330),'r');
plot(1:10:330,para,'ro','MarkerFaceColor','r','MarkerSize',8);
set(gca,'XTick',1:10:330,'XTickLabel',seq(ind),'FontSize',12)
ylabel('Scaled optimized parameters');
set(gca,'XTickLabelRotation',90)
% ylim([0 3.5]);
xlim([0 323]);
legend('ID135','location','best')
box on;grid on;hold off

end
%   End of HumanLiverRegen_para_optimizaton


%% ODE Function to solve
function dxdt = LiverRegenODE(t,x,k,x0)
global eps

% Set variable names
G = x(11);
Q = max(0,x(1)); P = max(0,x(2)); R = max(0,x(3)); 
N = (Q+G*(P+R))+eps; % Calculate N = Q + G(P+R) + epsilon
IE = x(4); GF = max(1,x(5)); ECM = x(6); TNF = x(7); JAK = x(8); STAT = x(9); SOCS = x(10);
IE0 = x0(4); GF0 = x0(5); ECM0 = x0(6); TNF0 = x0(7); JAK0 = x0(8); STAT0 = x0(9); SOCS0 = x0(10);

% Calculate sigmas
sr = 0.5*(1+tanh((k(29)-GF)/k(30))); % sigma_req
sa = 0.5*(1+tanh((k(32)-((N)/k(2)))/k(33))); % sigma_ap

% Calculate reaction rates
r1 = k(24)*(IE-IE0)*Q; % Priming of Quiescent cells
r2 = k(26)*ECM*R; % Replicating cell returning to Quiescence
r3 = k(28)*sr*P; % Requiescence of Primed cells
r4 = k(25)*(GF-GF0)*P; % Primed cells begining Replication
r5 = k(27)*R; % Doubling of Replicating cells
ra1 = k(31)*sa*Q; % cell death of Quiescent cells
ra2 = k(31)*sa*P; % cell death of Primed cells
ra3 = k(31)*sa*R; % cell death of Replicating cells

r6 = k(3)*k(2)/(N); % TNF production by stimulus
r7 = k(5)*TNF/(TNF+k(6)); % TNF production of JAK
r8 = k(4)*TNF; % TNF degradation
r9 = k(7)*JAK; % JAK degredation
r10 = k(9)*JAK*k(8)^2/(k(8)^2+k(10)*(1+SOCS/k(15))); % STAT3 production by JAK
r11 = k(16)*STAT/(STAT+k(17)); % IE production by STAT3
r12 = k(12)*STAT/(STAT+k(13)); % SOCS3 production by STAT3
r13 = k(11)*STAT; % STAT3 degredation
r14 = k(14)*SOCS; % SOCS3 degredation
r15 = k(18)*IE; % IE degredation
r16 = k(19)*TNF*ECM; % ECM degredation by TNF activated MMPs
r17 = k(20)*ECM; % ECM degredation
r18 = k(21)*k(2)/N; % GF production by stimulus
r19 = k(23)*GF*ECM; % GF uptake/production by ECM
r20 = k(22)*GF; % GF degredation

% Set differential equations
dxdt(1) = -r1 + r2 + r3 - ra1; % Q Phase
dxdt(2) = r1 - r4 - r3 - ra2; % P Phase
dxdt(3) = r4 - r2 + r5 - ra3; % R Phase

dxdt(4) = r11 - r15 + k(38); % IE genes
dxdt(5) = r18 - r19 - r20 + k(40); % GF
dxdt(6) = -r16 - r17 + k(39); % ECM

dxdt(7) = r6 - r7 - r8 + k(34); % TNF
dxdt(8) = r7 - r9 + k(35); % JAK
dxdt(9) = r10 - r11 - r12 - r13 + k(36); % STAT3
dxdt(10) = r12 - r14 + k(37); % SOCS3

dxdt(11) = (k(2)/(N))*k(41) - k(2)*k(41); % Cell Mass (M)

dxdt = dxdt';
end
%   End of LiverRegenODE


%% error calculation function for fmincon
function [err] = error_cal(k0)
global data data_t x0 tStart tEnd k lamda1 lamda2 k0_ch W eps Nss

k(2:33) = k0(1:32);
k(41) = k0(33);

% Steady-State parameters
k(34) = -k(3)*k(2) + k(5)/(1+k(6)) + k(4); %-22.3; % TNF production
k(35) = -k(5)/(1+k(6)) + k(7); %-1.6; % JAK production
k(36) = -k(9)*k(8)^2/(k(8)^2+k(10)*(1+1/k(15))) + ...
    k(16)/(1+k(17)) + k(12)/(1+k(13)) + k(11); %2.39e4; % STAT3 production
k(37) = -k(12)/(1+k(13)) + k(14); %2.4e4; % SOCS production
k(38) = -k(16)/(1+k(17)) + k(18); %-8.16; % IE gene production
k(39) = k(19) + k(20); %40; % ECM production
k(40) = -k(21)*k(2)/(Nss+eps) + k(23) + k(22); %-1.6; % GF production

%solving the suystem of equations
[t,xf] = ode15s(@(t,x)LiverRegenODE(t,x,k,x0), [tStart:.1:tEnd], x0);

% Remove negative values not used in calculations
xf(xf(:,1) < 0,1) = 0;
xf(xf(:,2) < 0,2) = 0;
xf(xf(:,3) < 0,3) = 0;

%calculationg each term of objective function
Sim_value = zeros(1,length(data_t));
Nf = xf(:,1)+xf(:,11).*(xf(:,2)+xf(:,3));

%finding simualted time point corresponding to the data time point
for i = 1:length(data_t)
    j = find(t == 24*data_t(i),1);
    Sim_value(1,i) = Nf(j,1);
end

%difference in simulated and data value of the output
e = Sim_value-data;

%error calculation with regularization using Ridge and Lasso
err = e*W*e'+lamda1*norm(((k0./k0_ch)-1),1)+lamda2*(((k0./k0_ch)-1)*((k0./k0_ch)-1)');

end
%   end of error_cal
